from datetime import datetime, timedelta
from typing import Dict, Type

import csp
import orjson
import pytest
from csp import ts

from csp_gateway import (
    AddChannelsToGraphOutput,
    Channels,
    Gateway,
    GatewayChannels,
    GatewayModule,
    GatewayStruct,
    ReadWriteMode,
    ReplayEngineJSON,
    State,
)
from csp_gateway.server.gateway.csp.channels import _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD
from csp_gateway.testing.shared_helpful_classes import (
    MyGateway,
    MyGatewayChannels,
    MySetModule,
    MyStruct,
)


class MyStrStruct(GatewayStruct):
    foo: str


class MySmallGatewayChannels(GatewayChannels):
    example: ts[int] = None
    struct_with_str: ts[MyStrStruct] = None
    s_example: ts[State[int]] = None
    my_str_basket: Dict[str, ts[float]] = None


class MySmallGateway(Gateway):
    channels_model: Type[Channels] = MySmallGatewayChannels


class MySetStrStructModule(GatewayModule):
    my_data: ts[MyStrStruct]

    def connect(self, channels: MySmallGatewayChannels) -> None:
        channels.set_channel(MySmallGatewayChannels.struct_with_str, self.my_data)


class MyReplayEngineJSON(ReplayEngineJSON):
    def dynamic_keys(self):
        return {MyGatewayChannels.my_str_basket: ["my_key", "my_key2"]}


class MyModuleTestSameTimeDifferentCycle(GatewayModule):
    my_data: ts[MyStruct]
    my_data2: ts[MyStruct]

    def connect(self, channels: MyGatewayChannels) -> None:
        # Build a set of inputs based on the test cases
        channels.set_channel(MyGatewayChannels.my_channel, self.my_data)
        my_list_data = csp.collect([self.my_data, self.my_data2])
        channels.set_channel(
            MyGatewayChannels.my_list_channel,
            my_list_data,
        )


@pytest.mark.parametrize("by_key", [True, False])
def test_json_module_output(tmpdir, by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    file = str(tmpdir.join("json_test_data.json"))
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    json_module = ReplayEngineJSON(
        selection=channels,
        filename=file,
        read_write_mode=ReadWriteMode.WRITE,
    )
    gateway_writing = MyGateway(modules=[setter, json_module, setter, setter], channels=MyGatewayChannels())
    csp.run(gateway_writing.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))

    json_module_read = ReplayEngineJSON(
        selection=channels,
        filename=file,
        read_write_mode=ReadWriteMode.READ,
        start_writing=datetime(2099, 1, 1),  # far in the future to never stop reading
    )
    gateway_reading = MyGateway(
        modules=[
            setter,
            json_module_read,
            AddChannelsToGraphOutput(selection=channels),
        ],
        channels=MyGatewayChannels(),
    )
    out = csp.run(gateway_reading.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    # TODO: encode/decode Numpy1DArray
    assert len(out["my_channel"]) == 4
    assert len(out["my_list_channel"]) == 4
    assert len(out["my_list_channel"][0][1]) == 2
    assert len(out["my_list_channel"][1][1]) == 2
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 4
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 4
    assert len(out["my_str_basket[my_key]"]) == 4
    assert len(out["my_str_basket[my_key2]"]) == 4
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 4
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 4


@pytest.mark.parametrize("by_key", [True, False])
def test_json_module_nested_directory(tmpdir, by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    file = str(tmpdir.join("/non_existent_dir/json_test_data.json"))
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    json_module = ReplayEngineJSON(
        selection=channels,
        filename=file,
        read_write_mode=ReadWriteMode.WRITE,
    )
    gateway_writing = MyGateway(modules=[setter, json_module, setter, setter], channels=MyGatewayChannels())
    csp.run(gateway_writing.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))

    json_module_read = ReplayEngineJSON(
        selection=channels,
        filename=file,
        read_write_mode=ReadWriteMode.READ,
        start_writing=datetime(2099, 1, 1),  # far in the future to never stop reading
    )
    gateway_reading = MyGateway(
        modules=[
            setter,
            json_module_read,
            AddChannelsToGraphOutput(selection=channels),
        ],
        channels=MyGatewayChannels(),
    )
    out = csp.run(gateway_reading.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    # TODO: encode/decode Numpy1DArray
    assert len(out["my_channel"]) == 4
    assert len(out["my_list_channel"]) == 4
    assert len(out["my_list_channel"][0][1]) == 2
    assert len(out["my_list_channel"][1][1]) == 2
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 4
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 4
    assert len(out["my_str_basket[my_key]"]) == 4
    assert len(out["my_str_basket[my_key2]"]) == 4
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 4
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 4


@pytest.mark.parametrize("by_key", [True, False])
def test_custom_json_module_output(tmpdir, by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    file = str(tmpdir.join("json_test_data.json"))
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    json_module = ReplayEngineJSON(
        selection=channels,
        filename=file,
        read_write_mode=ReadWriteMode.WRITE,
    )
    gateway_writing = MyGateway(modules=[setter, json_module, setter, setter], channels=MyGatewayChannels())
    csp.run(gateway_writing.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))

    # we can make our own custom ReplayEngineJSON class that specifies dynamic keys
    json_module_read = MyReplayEngineJSON(
        selection=channels,
        filename=file,
        read_write_mode=ReadWriteMode.READ,
        start_writing=datetime(2099, 1, 1),  # far in the future to never stop reading
    )
    gateway_reading = MyGateway(
        modules=[json_module_read, AddChannelsToGraphOutput(selection=channels)],
        channels=MyGatewayChannels(),
    )
    out = csp.run(gateway_reading.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    # TODO: encode/decode Numpy1DArray
    assert len(out["my_channel"]) == 3
    assert len(out["my_list_channel"]) == 3
    assert len(out["my_list_channel"][0][1]) == 2
    assert len(out["my_list_channel"][1][1]) == 2
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 3
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 3
    assert len(out["my_str_basket[my_key]"]) == 3
    assert len(out["my_str_basket[my_key2]"]) == 3
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 3
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 3


def test_json_module_output_diff_timestamp_tick(tmpdir):
    setter = MyModuleTestSameTimeDifferentCycle(
        my_data=csp.curve(
            MyStruct,
            [(timedelta(), MyStruct(foo=1.0)), (timedelta(), MyStruct(foo=3.0))],
        ),
        my_data2=csp.curve(
            MyStruct,
            [
                (timedelta(), MyStruct(foo=2.0)),
            ],
        ),
    )
    file = str(tmpdir.join("json_test_data.json"))
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
    ]
    json_module = ReplayEngineJSON(
        selection=channels,
        filename=file,
        read_write_mode=ReadWriteMode.WRITE,
    )
    gateway_writing = MyGateway(modules=[setter, json_module], channels=MyGatewayChannels())
    csp.run(gateway_writing.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))

    json_module_read = ReplayEngineJSON(
        selection=channels,
        filename=file,
        read_write_mode=ReadWriteMode.READ,
        start_writing=datetime(2099, 1, 1),  # far in the future to never stop reading
    )
    gateway_reading = MyGateway(
        modules=[json_module_read, AddChannelsToGraphOutput(selection=channels)],
        channels=MyGatewayChannels(),
    )
    out = csp.run(gateway_reading.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    # TODO: encode/decode Numpy1DArray
    assert len(out["my_channel"]) == 2
    assert len(out["my_list_channel"]) == 2
    assert len(out["my_list_channel"][0][1]) == 2
    # on the second tick of my_data, my_data2 does not
    # tick so our list is just of length 1
    assert len(out["my_list_channel"][1][1]) == 1
    assert out["my_list_channel"][1][1][0].foo == 3.0


def test_json_module_delimiter(tmpdir):
    file = str(tmpdir.join("json_test_data.json"))
    setter = MySetStrStructModule(my_data=csp.const(MyStrStruct(foo="\nis\nthe\ndelimiter\n")))
    json_module = ReplayEngineJSON(
        selection=[MySmallGatewayChannels.struct_with_str],
        filename=file,
        read_write_mode=ReadWriteMode.WRITE,
    )
    gateway_writing = MySmallGateway(modules=[setter, json_module], channels=MySmallGatewayChannels())
    csp.run(gateway_writing.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    json_module_read = ReplayEngineJSON(
        selection=[MySmallGatewayChannels.struct_with_str],
        filename=file,
        read_write_mode=ReadWriteMode.READ,
        start_writing=datetime(2099, 1, 1),  # far in the future to never stop reading
    )
    output_module = AddChannelsToGraphOutput(selection=[MySmallGatewayChannels.struct_with_str])
    gateway_reading = MySmallGateway(modules=[json_module_read, output_module], channels=MySmallGatewayChannels())
    out = csp.run(gateway_reading.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert out["struct_with_str"][0][1].foo == "\nis\nthe\ndelimiter\n"
    assert len(out["struct_with_str"]) == 1


@pytest.mark.parametrize("exclude", [True, False])
def test_json_module_all_channels(tmpdir, exclude):
    file = str(tmpdir.join("json_test_data.json"))
    my_struct = MyStrStruct(foo="|is|the|delimiter|", id="test")
    setter = MySetStrStructModule(my_data=csp.const(my_struct))
    json_module = ReplayEngineJSON(
        filename=file,
        read_write_mode=ReadWriteMode.WRITE,
    )
    gateway_writing = MySmallGateway(modules=[setter, json_module], channels=MySmallGatewayChannels())
    csp.run(gateway_writing.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    json_module_read = ReplayEngineJSON(
        selection=[MySmallGatewayChannels.struct_with_str],
        filename=file,
        read_write_mode=ReadWriteMode.READ,
        start_writing=datetime(2099, 1, 1),  # far in the future to never stop reading
    )
    if exclude:
        json_module_read.subscribe_with_struct_id = False
        json_module_read.subscribe_with_struct_timestamp = False
    output_module = AddChannelsToGraphOutput(selection=[MySmallGatewayChannels.struct_with_str])
    gateway_reading = MySmallGateway(modules=[json_module_read, output_module], channels=MySmallGatewayChannels())
    out = csp.run(gateway_reading.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert out["struct_with_str"][0][1].foo == "|is|the|delimiter|"
    assert len(out["struct_with_str"]) == 1
    if exclude:
        assert out["struct_with_str"][0][1].id != my_struct.id
        assert out["struct_with_str"][0][1].timestamp != my_struct.timestamp
    else:
        assert out["struct_with_str"][0][1].id == my_struct.id
        assert out["struct_with_str"][0][1].timestamp == my_struct.timestamp


@pytest.mark.parametrize("writing_start_type", ["datetime", "timedelta"])
def test_switch_automatically_to_write(writing_start_type, tmpdir):
    if writing_start_type == "datetime":
        start_writing = datetime(2020, 12, 12)
    elif writing_start_type == "timedelta":
        start_writing = timedelta(days=364)
    file = str(tmpdir.join("json_test_data.json"))
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    json_module = ReplayEngineJSON(
        selection=channels,
        filename=file,
        read_write_mode=ReadWriteMode.WRITE,
    )
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    gateway = MyGateway(modules=[json_module, setter], channels=MyGatewayChannels())
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1))
    with open(file, "r") as json_file:
        assert len(list(json_file)) == 1

    json_module_read_and_write = ReplayEngineJSON(selection=channels, filename=file, start_writing=start_writing)
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0), timedelta(days=365)),
        my_data2=csp.const(MyStruct(foo=2.0), timedelta(days=365)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)], timedelta(days=365)),
    )
    new_gateway = MyGateway(modules=[json_module_read_and_write, setter], channels=MyGatewayChannels())
    csp.run(new_gateway.graph, starttime=datetime(2020, 1, 1))
    with open(file, "r") as json_file:
        items = list(json_file)
        assert len(items) == 2
        json_dict_0 = orjson.loads(items[0])
        first_date = datetime.fromisoformat(json_dict_0[_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD]).replace(tzinfo=None)
        assert first_date == datetime(2020, 1, 1)

        json_dict_1 = orjson.loads(items[1])
        second_date = datetime.fromisoformat(json_dict_1[_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD]).replace(tzinfo=None)
        assert second_date == (datetime(2020, 1, 1) + timedelta(days=365))


@pytest.mark.parametrize("read_write_mode", [ReadWriteMode.READ_AND_WRITE, ReadWriteMode.READ])
def test_overwrite_json(read_write_mode, tmpdir):
    file = str(tmpdir.join("json_test_data.json"))
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    setup_json_module = ReplayEngineJSON(
        selection=channels,
        filename=file,
        read_write_mode=ReadWriteMode.WRITE,
    )
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    gateway = MyGateway(modules=[setup_json_module, setter], channels=MyGatewayChannels())
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1))
    with open(file, "r") as json_file:
        assert len(list(json_file)) == 1

    json_module = ReplayEngineJSON(
        selection=channels,
        filename=file,
        start_writing=datetime(2020, 12, 12),
        read_write_mode=read_write_mode,
        overwrite_if_writing=True,
    )
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0), timedelta(days=365)),
        my_data2=csp.const(MyStruct(foo=2.0), timedelta(days=365)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)], timedelta(days=365)),
    )
    new_gateway = MyGateway(modules=[json_module, setter], channels=MyGatewayChannels())
    csp.run(new_gateway.graph, starttime=datetime(2020, 1, 1))
    with open(file, "r") as json_file:
        items = list(json_file)
        assert len(items) == 1
        json_dict = orjson.loads(items[0])
        actual_date = datetime.fromisoformat(json_dict[_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD]).replace(tzinfo=None)

        if read_write_mode is ReadWriteMode.READ_AND_WRITE:
            assert actual_date == datetime(2020, 1, 1) + timedelta(days=365)
        elif read_write_mode is ReadWriteMode.READ:
            assert actual_date == datetime(2020, 1, 1)
        else:
            raise TypeError(f"read_write_mode should be a ReadWriteMode enum, got {read_write_mode}")



from datetime import datetime

import orjson
from csp import ts
from csp.impl.pulladapter import PullInputAdapter
from csp.impl.wiring import py_pull_adapter_def

from csp_gateway.server.gateway.csp.channels import _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD

__all__ = ("JSONPullAdapter",)


# The Impl object is created at runtime when the graph is converted into the runtime engine
# it does not exist at graph building time!
class JSONPullAdapterImpl(PullInputAdapter):
    def __init__(self, filename: str):
        self._filename = filename
        self._file = None
        self._next_row = None
        super().__init__()

    def start(self, start_time, end_time):
        self._file = open(self._filename, "r")
        for line in self._file:
            self._next_row = line
            json_dict = orjson.loads(line)
            time_str = json_dict[_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD]
            # since csp engine times are in utc we can just
            # drop the timezone info to compare it to start_time
            time = datetime.fromisoformat(time_str).replace(tzinfo=None)
            if time >= start_time:
                break

        super().start(start_time, end_time)

    def stop(self):
        self._file.close()

    def next(self):
        if self._next_row is None:
            return None

        while True:
            json_dict = orjson.loads(self._next_row)
            time_str = json_dict[_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD]
            time = datetime.fromisoformat(time_str).replace(tzinfo=None)
            try:
                self._next_row = next(self._file)
            except StopIteration:
                self._next_row = None
            return time, json_dict


# MyPullAdapter is the graph-building time construct.  This is simply a representation of what the
# input adapter is and how to create it, including the Impl to use and arguments to pass into it upon construction
JSONPullAdapter = py_pull_adapter_def("JSONPullAdapter", JSONPullAdapterImpl, ts[dict], filename=str)



import itertools
from datetime import datetime, timedelta
from typing import Dict, Optional, Type

import csp
import pytest
from csp import ts

from csp_gateway import Channels, ChannelSelection, Gateway, GatewayChannels, GatewayModule, GatewayStruct


class MyStruct(GatewayStruct):
    foo: float


class MyGatewayChannels(GatewayChannels):
    my_str_basket: Dict[str, ts[MyStruct]] = None
    my_channel1: ts[MyStruct] = None
    my_channel2: ts[MyStruct] = None


class MyGateway(Gateway):
    channels_model: Type[Channels] = MyGatewayChannels  # type: ignore[assignment]


class MySetModuleFeedback(GatewayModule):
    my_data: ts[MyStruct]
    my_set_channel: str
    my_get_channel: str
    filter_ticks: int = 3

    def connect(self, channels: MyGatewayChannels) -> None:
        # Build a set of inputs based on the test cases
        input = channels.get_channel(self.my_get_channel)
        output = csp.merge(self.my_data, input)
        # We only tick output if input has only ticked once or less
        flag = csp.default(csp.count(input) < csp.const(self.filter_ticks), True)
        res = csp.filter(flag, output)
        channels.set_channel(self.my_set_channel, res)


class MySetModuleDict(GatewayModule):
    my_data: ts[MyStruct]
    my_dep: Optional[str] = None
    my_key: str
    by_key: bool = True
    filter_ticks: int = 2

    def dynamic_keys(self):
        return {MyGatewayChannels.my_str_basket: [self.my_key]}

    def connect(self, channels: MyGatewayChannels) -> None:
        # Build a set of inputs based on the test cases
        output = self.my_data
        input = None
        if self.my_dep is not None:
            input = channels.get_channel(MyGatewayChannels.my_str_basket, self.my_dep)
            output = csp.merge(output, input)
        if input is not None:
            flag = csp.default(csp.count(input) < csp.const(self.filter_ticks), True)
            output = csp.filter(flag, output)
        if self.by_key:
            channels.set_channel(MyGatewayChannels.my_str_basket, output, self.my_key)

        else:
            channels.set_channel(
                MyGatewayChannels.my_str_basket,
                {self.my_key: output},
            )


class MyGetModule(GatewayModule):
    requires: Optional[ChannelSelection] = []  # so that we dont error if a channel isnt provided

    def connect(self, channels: MyGatewayChannels) -> None:
        for k, v in channels.get_channel(MyGatewayChannels.my_str_basket).items():
            csp.add_graph_output(f"my_str_basket[{k}]", v)
        csp.add_graph_output("my_channel2", channels.get_channel(MyGatewayChannels.my_channel2))
        csp.add_graph_output("my_channel1", channels.get_channel(MyGatewayChannels.my_channel1))


@pytest.mark.parametrize("ordering", list(itertools.permutations([0, 1, 2])))
def test_set_get(ordering):
    """Test that two modules can get/set different keys of the same dict channel"""
    setter = MySetModuleDict(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_key="my_key",
    )
    setter2 = MySetModuleDict(
        my_data=csp.const(MyStruct(foo=2.0)),
        my_key="my_key2",
        my_dep="my_key",
    )
    getter = MyGetModule()
    base_order = [setter, setter2, getter]
    modules = [base_order[i] for i in ordering]
    gateway = MyGateway(modules=modules, channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert "my_str_basket[my_key]" in out
    assert len(out["my_str_basket[my_key]"]) == 1
    assert "my_str_basket[my_key2]" in out
    # This should be 1. We send in the data at the same time
    # Since csp.merge drops the right tick if both inputs tick
    # at the same time, we should actually get just 1 tick.
    assert len(out["my_str_basket[my_key2]"]) == 1


def test_set_get_loop_same_module():
    """Test that two modules can get/set the same channel (forms a cycle)."""
    setter = MySetModuleFeedback(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_set_channel=MyGatewayChannels.my_channel1,
        my_get_channel=MyGatewayChannels.my_channel1,
    )
    getter = MyGetModule()
    gateway = MyGateway(modules=[setter, getter], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert "my_channel1" in out

    # We add a feedback since we now formed a cycle with the same channel
    # Since we filter out the ouput in our setter if the input ticks 3 times
    # (defined in the 'connect' method of MySetModuleFeedback)
    # We get 3 ticks here
    assert len(out["my_channel1"]) == 3


@pytest.mark.parametrize("ordering", list(itertools.permutations([0, 1, 2])))
def test_set_get_loop_across_modules(ordering):
    """Test that two modules can get/set different channels that form a cycle."""
    setter = MySetModuleFeedback(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_set_channel=MyGatewayChannels.my_channel1,
        my_get_channel=MyGatewayChannels.my_channel2,
    )
    setter2 = MySetModuleFeedback(
        my_data=csp.const(MyStruct(foo=2.0)),
        my_set_channel=MyGatewayChannels.my_channel2,
        my_get_channel=MyGatewayChannels.my_channel1,
    )
    getter = MyGetModule()
    base_order = [setter, setter2, getter]
    modules = [base_order[i] for i in ordering]
    gateway = MyGateway(modules=modules, channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert "my_channel1" in out
    assert "my_channel2" in out

    # We need to add a feedback since we now formed a cycle with channels across modules.
    # However, we only need the feedback on one edge.
    # We add feedbacks, as needed, in the order that modules are listed in the "modules" list
    # Thus, whichever module that is added first, it's "set_channel" call, which forms
    # a cycle, will get a feedback added. The addition of this feedback means that the
    # cycle is broken, so the other module will not get a new feedback.
    set_my_channel1_first = ordering.index(0) > ordering.index(1)
    if set_my_channel1_first:
        # The module that calls "set_channel" on "MyGatewayChannels.my_channel1"
        # is processed first. This means, we add the feedback right before
        # we call set. Thus, this input is delayed, so by the time it's
        # second tick comes, the other channel "MyGatewayChannels.my_channel2" has already ticked
        # so we filter out the output.
        assert len(out["my_channel1"]) == 2
        assert len(out["my_channel2"]) == 3
    else:
        # Similar logic as above, but flip the channels
        assert len(out["my_channel1"]) == 3
        assert len(out["my_channel2"]) == 2


@pytest.mark.parametrize("ordering", list(itertools.permutations([0, 1, 2])))
@pytest.mark.parametrize("by_key", [True, False])
def test_set_get_loop_dict(by_key, ordering):
    """Test that two modules can get/set different keys of the same dict channel that form a cycle."""
    setter = MySetModuleDict(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_key="my_key",
        my_dep="my_key2",
        by_key=by_key,
    )
    setter2 = MySetModuleDict(
        my_data=csp.const(MyStruct(foo=2.0)),
        my_key="my_key2",
        my_dep="my_key",
        by_key=by_key,
    )
    getter = MyGetModule()
    base_order = [setter, setter2, getter]
    modules = [base_order[i] for i in ordering]
    gateway = MyGateway(modules=modules, channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert "my_str_basket[my_key]" in out
    assert "my_str_basket[my_key2]" in out

    # We need to add a feedback since we now formed a cycle with different dict keys
    # across modules. However, we only need the feedback on one edge.
    # We add feedbacks, as needed, in the order that modules are listed in the "modules" list
    # Thus, whichever module that is added first, it's "set_channel" call, which forms
    # a cycle, will get a feedback added. The addition of this feedback means that the
    # cycle is broken, so the other module will not get a new feedback.
    set_my_key_edge_first = ordering.index(0) > ordering.index(1)
    if set_my_key_edge_first:
        # The module that calls "set_channel" on key "my_key"
        # is processed first. This means, we add the feedback right before
        # we call set. Thus, this input is delayed, so by the time it's
        # second tick comes, the other key has already ticked so we filter
        # out the output.
        assert len(out["my_str_basket[my_key]"]) == 1
        assert len(out["my_str_basket[my_key2]"]) == 2
    else:
        # Similar logic as above, but flip "my_key" with "my_key2"
        assert len(out["my_str_basket[my_key]"]) == 2
        assert len(out["my_str_basket[my_key2]"]) == 1



from logging import INFO, basicConfig
from pathlib import Path

from csp_gateway import (
    Gateway,
    GatewaySettings,
    MountOutputsFolder,
    MountPerspectiveTables,
    MountRestRoutes,
)
from csp_gateway.server.config import load_gateway

from .simple import ExampleGatewayChannels, ExampleModule

# csp-gateway is configured as a hydra application,
# but it can also be instantiated directly as we do so here:

# instantiate gateway
gateway = Gateway(
    settings=GatewaySettings(),
    modules=[
        ExampleModule(),
        MountOutputsFolder(),
        MountPerspectiveTables(),
        MountRestRoutes(force_mount_all=True),
    ],
    channels=ExampleGatewayChannels(),
)

if __name__ == "__main__":
    # To run, we could run our object directly:
    # gateway.start(rest=True, ui=True)

    # But instead, lets run the same code via hydra
    # We can use our own custom config, in config/demo.yaml
    # which inherits from csp-gateway's example config.
    #
    # With hydra, we can easily construct hierarchichal,
    # extensible configurations for all our modules!
    gateway = load_gateway(
        overrides=["+config=demo"],
        config_dir=Path(__file__).parent,
    )

    # Set our log level to info. If using hydra,
    # we have even more configuration options at our disposal
    basicConfig(level=INFO)

    # Start the gateway
    gateway.start(rest=True, ui=True)

    # You can also run this directly via cli
    # > pip install csp-gateway
    # > csp-gateway-start --config-dir=csp_gateway/server/demo +config=demo

    # For more a more complicated example, see `./omnibus.py`



import logging
from pprint import pprint

import hydra
from ccflow import ModelRegistry

from csp_gateway import __version__

log = logging.getLogger(__name__)


@hydra.main(config_path="config", config_name="base", version_base=None)
def main(cfg):
    log.info("Loading csp-gateway config...")
    registry = ModelRegistry.root()
    registry.load_config(cfg=cfg, overwrite=True)
    gateway = registry["gateway"]

    log.info(f"Starting csp_gateway version {__version__}")
    kwargs = cfg["start"]
    if kwargs:  # i.e. start=False override on command line
        log.info(f"Starting gateway with arguments: {kwargs}")
        gateway.start(**kwargs)
    else:
        pprint(gateway.model_dump(by_alias=True))


if __name__ == "__main__":
    main()



from datetime import datetime, timedelta, timezone

import csp
import numpy as np
import orjson
from csp import ts

from csp_gateway import Gateway, GatewayChannels, GatewayModule, GatewayStruct
from csp_gateway.server import JSONPullAdapter
from csp_gateway.server.gateway.csp.channels import _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD
from csp_gateway.testing.harness import GatewayTestHarness
from csp_gateway.testing.shared_helpful_classes import MyEnum


class GWC(GatewayChannels):
    test_json: ts[dict] = None


class MyTestJSONGatewayModule(GatewayModule):
    filename: str

    def connect(self, channels: GWC):
        json_pull_adapter = JSONPullAdapter(self.filename)
        channels.set_channel(GWC.test_json, json_pull_adapter)


def _assert_equal(exp):
    def func(x):
        assert x == exp

    return func


def _write_to_jsonl(list_of_dicts, file):
    with open(file, "w") as test_file:
        for val in list_of_dicts:
            json_str = orjson.dumps(val, option=orjson.OPT_NAIVE_UTC | orjson.OPT_SERIALIZE_NUMPY).decode()
            test_file.write(json_str)
            test_file.write("\n")


def test_json_reading(tmpdir):
    test_json_values = [
        {_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1)},
        {
            _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 8),
            "float": 7.0,
            "str": "\nabc",
        },
    ]
    file = str(tmpdir.join("test_json_reading.json"))
    _write_to_jsonl(test_json_values, file)

    h = GatewayTestHarness(test_channels=[GWC.test_json])
    first_tick_value = (
        datetime(2020, 1, 1),
        {_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1, tzinfo=timezone.utc).isoformat()},
    )
    h.assert_ticked_values(GWC.test_json, _assert_equal([first_tick_value]))
    h.delay(timedelta(days=7))

    second_tick_value = (
        datetime(2020, 1, 8),
        {
            _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 8, tzinfo=timezone.utc).isoformat(),
            "float": 7.0,
            "str": "\nabc",
        },
    )
    h.assert_ticked_values(GWC.test_json, _assert_equal([first_tick_value, second_tick_value]))

    json_input_adapter_module = MyTestJSONGatewayModule(filename=file)
    gateway = Gateway(modules=[h, json_input_adapter_module], channels=GWC())
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1))


def test_first_entries_before_start(tmpdir):
    test_json_values = [
        {_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1)},
        {
            _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 2),
            "float": 7.0,
            "str": "\nabc",
        },
        {_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 3)},
    ]
    file = str(tmpdir.join("test_first_entries_before_start.json"))
    _write_to_jsonl(test_json_values, file)
    h = GatewayTestHarness(test_channels=[GWC.test_json])
    first_tick_value = (
        datetime(2020, 1, 3),
        {_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 3, tzinfo=timezone.utc).isoformat()},
    )
    h.assert_ticked_values(GWC.test_json, _assert_equal([first_tick_value]))
    json_input_adapter_module = MyTestJSONGatewayModule(filename=file)
    gateway = Gateway(modules=[h, json_input_adapter_module], channels=GWC())
    csp.run(gateway.graph, starttime=datetime(2020, 1, 3))


def test_first_tick_after_start(tmpdir):
    test_json_values = [
        {_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1)},
        {
            _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 2),
            "float": 7.0,
            "str": "\nabc",
        },
    ]

    file = str(tmpdir.join("test_first_tick_after_start.json"))
    _write_to_jsonl(test_json_values, file)

    h = GatewayTestHarness(test_channels=[GWC.test_json])
    h.assert_ticked(GWC.test_json, count=0)
    h.delay(timedelta(days=1))
    first_tick_value = (
        datetime(2020, 1, 1),
        {_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1, tzinfo=timezone.utc).isoformat()},
    )

    h.assert_ticked_values(GWC.test_json, _assert_equal([first_tick_value]))

    second_tick_value = (
        datetime(2020, 1, 2),
        {
            _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 2, tzinfo=timezone.utc).isoformat(),
            "float": 7.0,
            "str": "\nabc",
        },
    )
    h.delay(timedelta(days=1))
    h.assert_ticked_values(GWC.test_json, _assert_equal([first_tick_value, second_tick_value]))

    json_input_adapter_module = MyTestJSONGatewayModule(filename=file)
    gateway = Gateway(modules=[h, json_input_adapter_module], channels=GWC())
    csp.run(gateway.graph, starttime=datetime(2019, 12, 31))


def test_same_timestamp(tmpdir):
    test_json_values = [
        {_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1)},
        {
            _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1),
            "test_outer": 7.0,
        },
    ]
    file = str(tmpdir.join("test_same_timestamp.json"))
    _write_to_jsonl(test_json_values, file)

    h = GatewayTestHarness(test_channels=[GWC.test_json])
    ticked_values = [
        (
            datetime(2020, 1, 1),
            {_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1, tzinfo=timezone.utc).isoformat()},
        ),
        (
            datetime(2020, 1, 1),
            {
                _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1, tzinfo=timezone.utc).isoformat(),
                "test_outer": 7.0,
            },
        ),
    ]
    h.assert_ticked_values(GWC.test_json, _assert_equal([ticked_values[0]]))
    h.delay(timedelta())
    h.assert_ticked_values(GWC.test_json, _assert_equal(ticked_values))
    json_input_adapter_module = MyTestJSONGatewayModule(filename=file)
    gateway = Gateway(modules=[h, json_input_adapter_module], channels=GWC())
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1))


def test_gateway_struct(tmpdir):
    # GatewayStructs when automatically encoded
    # call the `to_dict()` method for us
    test_json_values = [
        {
            _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1),
            "test_outer": GatewayStruct(id="9", timestamp=datetime(2020, 1, 1)).to_dict(),
        },
    ]
    file = str(tmpdir.join("test_gateway_struct.json"))
    _write_to_jsonl(test_json_values, file)

    h = GatewayTestHarness(test_channels=[GWC.test_json])
    ticked_values = [
        (
            datetime(2020, 1, 1),
            {
                _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1, tzinfo=timezone.utc).isoformat(),
                "test_outer": {
                    "id": "9",
                    "timestamp": datetime(2020, 1, 1, tzinfo=timezone.utc).isoformat(),
                },
            },
        ),
    ]
    h.assert_ticked_values(GWC.test_json, _assert_equal(ticked_values))
    json_input_adapter_module = MyTestJSONGatewayModule(filename=file)
    gateway = Gateway(modules=[h, json_input_adapter_module], channels=GWC())
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1))


def test_enum(tmpdir):
    # Enums aren't JSON-serializable so we use their names instead
    test_json_values = [
        {
            _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1),
            "enum": MyEnum.ONE.name,
        },
    ]
    file = str(tmpdir.join("test_gateway_struct.json"))
    _write_to_jsonl(test_json_values, file)

    h = GatewayTestHarness(test_channels=[GWC.test_json])
    ticked_values = [
        (
            datetime(2020, 1, 1),
            {
                _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1, tzinfo=timezone.utc).isoformat(),
                "enum": MyEnum.ONE.name,
            },
        ),
    ]
    h.assert_ticked_values(GWC.test_json, _assert_equal(ticked_values))
    json_input_adapter_module = MyTestJSONGatewayModule(filename=file)
    gateway = Gateway(modules=[h, json_input_adapter_module], channels=GWC())
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1))


def test_numpy_array(tmpdir):
    test_json_values = [
        {
            _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD: datetime(2020, 1, 1),
            "np_array": np.array([5.0, 7.0]),
        },
    ]
    file = str(tmpdir.join("test_gateway_struct.json"))
    _write_to_jsonl(test_json_values, file)

    h = GatewayTestHarness(test_channels=[GWC.test_json])

    def assert_numpy_equal(x):
        assert x[0][0] == datetime(2020, 1, 1)
        assert x[0][1][_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD] == datetime(2020, 1, 1, tzinfo=timezone.utc).isoformat()
        assert np.array_equal(x[0][1]["np_array"], np.array([5.0, 7.0]))
        assert len(x) == 1
        assert len(x[0]) == 2
        assert len(x[0][1]) == 2

    h.assert_ticked_values(GWC.test_json, assert_numpy_equal)
    json_input_adapter_module = MyTestJSONGatewayModule(filename=file)
    gateway = Gateway(modules=[h, json_input_adapter_module], channels=GWC())
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1))



import logging
from datetime import datetime
from enum import Enum as PyEnum
from typing import Literal, Optional, Union

try:
    from csp.impl.enum import Enum as CspEnum, EnumMeta as CspEnumMeta
except ImportError:
    # If csp is not available, we can still use the basic types
    CspEnum = PyEnum
    CspEnumMeta = PyEnum
from pydantic import BaseModel, Field

log = logging.getLogger(__name__)

FilterWhere = Literal["==", "!=", "<", "<=", ">", ">="]
FilterWhereLambdaMap = {
    "==": lambda a, b: a == b,
    "!=": lambda a, b: a != b,
    "<": lambda a, b: a < b,
    "<=": lambda a, b: a <= b,
    ">": lambda a, b: a > b,
    ">=": lambda a, b: a >= b,
}


class FilterCondition(BaseModel):
    # in priority order
    value: Optional[Union[float, int, str]] = Field(None)
    # have to handle separately otherwise
    # all ints would be datetimes..
    when: Optional[datetime] = Field(None)
    attr: str = ""
    where: FilterWhere = "=="


class Filter(BaseModel):
    attr: str = ""
    by: FilterCondition

    def calculate(self, obj) -> bool:
        """
        calculate the filter condition on the object

        returns `True` if SHOULD NOT BE FILTERED, else `False`
        """
        try:
            if self.by.value is not None:
                lhs = getattr(obj, self.attr)
                # Convert enums attrs to strings during filtering
                if isinstance(lhs, (CspEnum, PyEnum, CspEnumMeta)):
                    lhs = lhs.name
                log.info(f"Filtering: {lhs} {self.by.where} {self.by.value}")
                return FilterWhereLambdaMap[self.by.where](lhs, self.by.value)
            if self.by.when is not None:
                log.info(f"Filtering: {getattr(obj, self.attr)} {self.by.where} {self.by.when}")
                return FilterWhereLambdaMap[self.by.where](getattr(obj, self.attr), self.by.when)
            if self.by.attr:
                log.info(f"Filtering: {getattr(obj, self.attr)} {self.by.where} {getattr(obj, self.by.attr)}")
                return FilterWhereLambdaMap[self.by.where](getattr(obj, self.attr), getattr(obj, self.by.attr))
        except (ValueError, AttributeError) as e:
            # TODO probably surface to webserver
            log.warning(f"Error during filtering {type(obj)} / {self}: {e}")

        # default case, if there was an issue assume its not included
        return False



import logging
from typing import Any, Dict, List, Optional, Set, Union, get_args, get_origin

from fastapi import APIRouter, HTTPException, Request
from pydantic import BaseModel

from csp_gateway.utils import NoProviderException

from ..utils import get_default_responses
from .shared import prepare_response

log = logging.getLogger(__name__)


__all__ = (
    "add_send_routes",
    "add_send_available_channels",
)


def add_send_routes(
    api_router: APIRouter,
    field: str,
    model: Union[BaseModel, List[BaseModel]] = None,
    subroute_key: Any = None,
) -> None:
    if model and get_origin(model) is list:
        is_list_model = True
        base_model = get_args(model)[0]
        list_model = model
    else:
        is_list_model = False
        base_model = model
        list_model = List[model]

    if subroute_key:

        async def send(key: subroute_key, data: Union[list_model, base_model], request: Request) -> list_model:  # type: ignore[valid-type]
            """
            Send data to a dictionary basket channel, where `key` is the key of the dictionary basket.
            If such a key does not exist or is not mounted, this endpoint will raise a `404` error.
            """

            log.debug(f"send: {data}")

            # Throw 404 if not a supported channel
            if not hasattr(request.app.gateway.channels, field):
                raise HTTPException(status_code=404, detail="Channel not found: {}".format(field))

            if not isinstance(data, list):
                data = [data]

            if len(data):
                # send to the edge
                try:
                    if is_list_model:
                        # send as a list
                        request.app.gateway.channels.send(
                            getattr(request.app.gateway.channels_model, field),
                            data,
                            key,
                        )
                    else:
                        # unroll and send individually
                        for data in data:
                            request.app.gateway.channels.send(
                                getattr(request.app.gateway.channels_model, field),
                                data,
                                key,
                            )
                except NoProviderException:
                    raise HTTPException(
                        status_code=404,
                        detail="Channel not found: {}/{}".format(field, key),
                    )

            # Emit the pydantic model back as it
            # will now have the `id` and `timestamp`
            return prepare_response(data, is_list_model=is_list_model)

        api_router.post(
            "/{}/{{key:path}}".format(field),
            responses=get_default_responses(),
            name="Send {} by key".format(field),
        )(send)
        api_router.post(
            "/{}/{{key:path}}".format(field.replace("_", "-")),
            responses=get_default_responses(),
            name="Send {} by key".format(field),
            include_in_schema=False,
        )(send)

        async def send(data: Dict[subroute_key, base_model], request: Request) -> Dict[subroute_key, base_model]:  # type: ignore[valid-type]
            log.debug(f"send: {data}")

            # Throw 404 if not a supported channel
            if not hasattr(request.app.gateway.channels, field):
                raise HTTPException(status_code=404, detail="Channel not found: {}".format(field))

            # send to the edge
            try:
                # send as a list
                request.app.gateway.channels.send(
                    getattr(request.app.gateway.channels_model, field),
                    data,
                )
            except NoProviderException:
                raise HTTPException(
                    status_code=404,
                    detail="Channel not found: {}".format(field),
                )

            # Emit the pydantic model back as it
            # will now have the `id` and `timestamp`
            return prepare_response(data, is_dict_basket=True)

        api_router.post(
            "/{}".format(field),
            responses=get_default_responses(),
            name="Send {}".format(field),
        )(send)
        api_router.post(
            "/{}".format(field.replace("_", "-")),
            responses=get_default_responses(),
            name="Send {}".format(field),
            include_in_schema=False,
        )(send)

    elif model:

        async def send(data: Union[list_model, base_model], request: Request) -> list_model:  # type: ignore[misc, valid-type]
            """
            Send data to a non-basket channel. This endpoint can accept either a single element, or a list of elements.
            A list of elements will be returned with `id` and `timestamp` fields assigned.
            Users should not provide `id` or `timestamp` fields as these will be ignored.
            """

            log.debug(f"send: {data}")

            # Throw 404 if not a supported channel
            if not hasattr(request.app.gateway.channels, field):
                raise HTTPException(status_code=404, detail="Channel not found: {}".format(field))

            if not isinstance(data, list):
                data = [data]

            if len(data):
                # send to the edge
                try:
                    if is_list_model:
                        # send as a list
                        request.app.gateway.channels.send(
                            getattr(request.app.gateway.channels_model, field),
                            data,
                        )
                    else:
                        # unroll and send individually
                        for data in data:
                            request.app.gateway.channels.send(
                                getattr(request.app.gateway.channels_model, field),
                                data,
                            )
                except NoProviderException:
                    raise HTTPException(
                        status_code=404,
                        detail="Channel not found: {}".format(field),
                    )

            # Emit the pydantic model back as it
            # will now have the `id` and `timestamp`
            return prepare_response(data, is_list_model=is_list_model)

        api_router.post(
            "/{}".format(field),
            responses=get_default_responses(),
            name="Send {}".format(field),
        )(send)
        api_router.post(
            "/{}".format(field.replace("_", "-")),
            responses=get_default_responses(),
            name="Send {}".format(field),
            include_in_schema=False,
        )(send)


def add_send_available_channels(api_router: APIRouter, fields: Optional[Set[str]] = None) -> None:
    @api_router.get(
        "/",
        responses=get_default_responses(),
        response_model=List[str],
    )
    async def get_send(request: Request) -> List[str]:
        """
        This endpoint will return a list of string values of all available channels under the `/send` route.
        """
        return sorted(
            field + ("" if indexer is None else f"/{indexer.name if hasattr(indexer, 'name') else indexer}")
            for field, indexer in request.app.gateway.channels._send_channels.keys()
            if fields is None or field in fields
        )



from typing import List, Optional, Type, Union

from fastapi import APIRouter, HTTPException, Request
from pydantic import BaseModel, Field

from csp_gateway.server import ChannelSelection, GatewayChannels, GatewayModule
from csp_gateway.server.web import GatewayWebApp, get_default_responses


class MountFieldRestRoutes(GatewayModule):
    """Mount rest routes for specific non-csp fields of the GatewayChannels.

    This is not done generically across all static fields as they may not always be serializable.
    """

    requires: Optional[ChannelSelection] = []
    fields: List[str] = Field(description="Static fields on the Channels that should be exposed via REST. These must be JSON serializable.")
    route: str = "/field"

    def connect(self, channels: GatewayChannels) -> None:
        # NO-OP
        ...

    def rest(self, app: GatewayWebApp) -> None:
        # Get API Router
        api_router: APIRouter = app.get_router("api")

        for field in self.fields:
            model = app.gateway.channels_model.get_outer_type(field)
            add_field_routes(api_router, field, self.route, model)

        @api_router.get(
            "{}".format(self.route),
            responses=get_default_responses(),
            response_model=List[str],
            include_in_schema=False,
        )
        async def get_field(request: Request) -> List[str]:
            """
            This endpoint will return a list of string values of all available channels under the `/field` route.
            """
            return self.fields


def add_field_routes(
    api_router: APIRouter,
    field: str,
    route: str,
    model: Union[BaseModel, Type],
) -> None:
    async def get_field(request: Request) -> model:  # type: ignore[misc, valid-type]
        """
        Get static field value on a static channel.
        """
        # Throw 404 if not a supported channel
        if not hasattr(request.app.gateway.channels, field):
            raise HTTPException(status_code=404, detail="Channel field not found: {}".format(field))

        # Grab the request off the edge
        try:
            res = getattr(request.app.gateway.channels, field)
        except AttributeError:
            raise HTTPException(
                status_code=404,
                detail="Channel field not found: {}".format(field),
            )

        return res

    api_router.get(
        "{}/{}".format(route, field),
        responses=get_default_responses(),
        response_model=model,
        name="Get Channel field {}".format(field),
    )(get_field)

    api_router.get(
        "/{}".format(field.replace("_", "-")),
        responses=get_default_responses(),
        response_model=model,
        include_in_schema=False,
    )(get_field)



import logging
import multiprocessing.pool
import os
import signal
from datetime import datetime, timedelta
from socket import gethostname
from time import sleep
from typing import Any, Callable, List, Optional, Type, Union, get_args, get_origin

import csp
from csp import ts
from pydantic import Field, PrivateAttr, create_model, model_validator

from csp_gateway.server.gateway import State
from csp_gateway.server.settings import Settings

from .csp import Channels, ChannelsFactory, ChannelsType, Module

__all__ = (
    "Gateway",
    "GatewayChannels",
    "GatewayModule",
)


MAX_END_TIME = datetime(2261, 12, 31, 23, 59, 50, 999999)
log = logging.getLogger(__name__)


class GatewayChannels(Channels):
    @classmethod
    def fields(cls) -> List[str]:
        return list(cls.model_fields.keys())


class GatewayModule(Module[GatewayChannels]):
    model_config = {
        "ignored_types": (
            csp.impl.wiring.GraphDefMeta,
            csp.impl.wiring.NodeDefMeta,
        ),
    }

    def __hash__(self):
        # So that csp doesn't complain about inability to memoize
        return id(self)

    def shutdown(self):
        """Perform cleanup when the module is shutting down"""
        return


class Gateway(ChannelsFactory[GatewayChannels]):
    """The Gateway defines a combination of channels and a collection of modules that get and send data to the channels.

    Furthermore, it builds the underling csp event processing graph, as well as the web application (for the REST, web sockets and graphical interfaces).
    It also contains settings to define the behavior of the web application and how the csp graph should be run.
    """

    # For pydantic 2
    model_config = {"ignored_types": (csp.impl.wiring.GraphDefMeta, csp.impl.wiring.NodeDefMeta), "arbitrary_types_allowed": True}

    modules: List[GatewayModule] = Field([], description="The list of modules that will operate on the channels.")

    # Substructures
    settings: Settings = Field(default_factory=Settings, description="Generic settings for the gateway")
    web_app: Any = Field(None, description="The gateway will populate this field with the web application handle once started.")
    channels_model: Type[GatewayChannels] = Field(
        default=GatewayChannels,
        description="The type of the channels. Users of a `Gateway` are expected to pass `channels`, and `channels_model` will"
        "be automatically inferred from the type. Developers can subclass `Gateway` and set the default value of"
        "`channels_model` to be the specific type of channels that users must provide.",
    )

    # Running attributes
    output: Any = Field(None, description="(Running attribute). The gateway will save the output of the csp.graph run once complete to this field")
    graph_built: bool = Field(
        False,
        description="(Running attribute). The gateway will set this field to True when the application has detected the csp graph build is complete",
    )
    graph_build_failed: bool = Field(
        False, description="(Running attribute). The gateway will set this field to True if the csp graph build has failed"
    )
    running: bool = Field(False, description="(Running attribute). The gateway will set this field to True once the csp graph is running")
    _in_test: bool = PrivateAttr(False)
    _module_shutdown_timeout: int = PrivateAttr()
    _dynamic_channels_instantiated: bool = PrivateAttr(False)

    def __init__(
        self,
        modules: List[Module[GatewayChannels]] = None,
        channels: ChannelsType = None,
        *args: str,
        **kwargs: str,
    ):
        log.info(f"Initializing Gateway - pid[{os.getpid()}]")
        channels = channels or GatewayChannels()
        # Note that the channels object passed into the init function here is not necessarily the channels object
        # that is used to build the graph. Things like dynamic channels can be added.
        super().__init__(modules=modules, channels=channels, *args, **kwargs)
        self.graph_built = False
        self.graph_build_failed = False
        self._in_test = False
        self._module_shutdown_timeout = 60

    def _instantiate_dynamic_channel(self, modules: List[Module[GatewayChannels]], channels: ChannelsType) -> GatewayChannels:
        if self._dynamic_channels_instantiated:
            return channels

        if modules is None:
            modules = []

        dynamic_channels = {}
        for m in modules:
            module_dynamic_channels = m.dynamic_channels() if hasattr(m, "dynamic_channels") else None
            if module_dynamic_channels:
                channels_with_state = m.dynamic_state_channels() if hasattr(m, "dynamic_state_channels") else None
                for n, t in module_dynamic_channels.items():
                    existing_type = dynamic_channels.get(n, None)
                    if existing_type is not None:
                        if t is not existing_type:
                            raise ValueError(f"Conflicting types for dynamic channel {n}.")

                    dynamic_channels[n] = t
                    if channels_with_state and n in channels_with_state:
                        if get_origin(t) is list:
                            t = get_args(t)[0]

                        dynamic_channels[f"s_{n}"] = State[t]

        if dynamic_channels:
            dynamic_channel_kwargs = {n: (ts[t], None) for n, t in dynamic_channels.items()}
            base_class = type(channels)
            new_channel_name = f"{base_class.__name__}WithDynamicChannels"
            channels_type = create_model(new_channel_name, __base__=base_class, **dynamic_channel_kwargs)
            channels = channels_type(**{f: getattr(channels, f) for f in channels.model_dump(exclude_defaults=True)})

        self._dynamic_channels_instantiated = True
        return channels

    @model_validator(mode="before")
    @classmethod
    def _model_validate(cls, values):
        """Root validator to append "user_modules" to list of modules."""
        values["modules"] = values.get("modules") or []
        values["modules"].extend(values.pop("user_modules", []))
        return values

    def __getattribute__(self, attr: str) -> Any:
        if attr in ("channels", "state"):
            if not self.running:
                raise Exception("Can only access `{}` when engine is running".format(attr))
        return object.__getattribute__(self, attr)

    @csp.graph
    def graph(self, user_graph: Callable[[GatewayChannels], Any] = None):  # type: ignore[no-untyped-def]
        """Generates the csp graph corresponding to the gateway application.

        This is the graph that will be called by the `start` method

        This function can be passed to a csp.run call directly, which is especially useful if you want to control the graph running
        yourself, perhaps to save the output values from csp.add_graph_output, or to customize the arguments passed
        to csp.run further (for profiling, etc).

        Args:
            user_graph: A function that will be called with the channels to augment the existing application graph.
        """
        # Detect csp stopped, call first so it executes its stop block first
        self._stop_csp_detector()

        try:
            self.channels = self.build(channels=self._instantiate_dynamic_channel(self.modules, object.__getattribute__(self, "channels")))

            self.graph_built = True

            if user_graph:
                user_graph_result = user_graph(object.__getattribute__(self, "channels"))
                if user_graph_result is None:
                    # add a placeholder to ensure exit does not use os._exit
                    user_graph_result = csp.const(True)
                csp.add_graph_output("user_graph", user_graph_result)

        except Exception:
            self.graph_build_failed = True
            raise

        log.info("Launching CSP")
        csp.log(logging.INFO, "CSP Running", csp.const(True), logger=log)

        # Detect csp started, call last so it executes its start block last
        self._start_csp_detector()

    @csp.node
    def _start_csp_detector(self):
        with csp.start():
            self.running = True

    @csp.node
    def _stop_csp_detector(self):
        with csp.stop():
            self.running = False

    def start(
        self,
        user_graph: Optional[Any] = None,
        realtime: bool = True,
        block: bool = True,
        show: bool = False,
        rest: bool = False,
        ui: bool = False,
        _in_test: bool = False,
        starttime: Optional[datetime] = None,
        endtime: Optional[Union[datetime, timedelta]] = None,
        build_timeout: int = 30,
        module_shutdown_timeout: int = 60,
        **uvicorn_kwargs: Any,
    ) -> None:
        """Starts the application corresponding to the Gateway.

        Depending on the provided settings, this will start the web application as well as run the csp graph.
        To return the csp graph independent of the web application, look at the `graph` method.

        Args:
            user_graph: A function that will be called with the channels to augment the existing application graph.
            realtime: Whether to run the csp graph in realtime mode
            block: Whether the csp graph should be run in the foreground or on a background thread. Selecting `rest=True`
                will force this setting to False.
            show: Whether to write the csp graph to file (tmp.png) and return without running.
            rest: Whether to launch the web application as part of the Gateway (i.e. for the REST endpoints). If `True`, will
                force `block=False`.
            ui: Whether to equip the web application with the pieces necessary to run the Perspective-based UI.
            starttime: Start time of the csp graph run.
            endtime: End time of the csp graph run.
            build_timeout: Timeout that the web application uses to wait for the csp graph to start running successfully in the background.
                If the csp graph is not running by the timeout, the web application will shut down.
            module_shutdown_timeout: Timeout that the shutdown method uses to wait for all the modules to shutdown
            uvicorn_kwargs: Additional kwargs to pass to the [uvicorn server config](https://www.uvicorn.org/settings/). Also see `GatewayWebApp`.
        """
        # to avoid hard shutdown, starting webserver, etc
        self._in_test = _in_test
        self._module_shutdown_timeout = module_shutdown_timeout

        try:
            if show:
                # Show graph and return without running
                os.makedirs("outputs", exist_ok=True)
                csp.show_graph(self.graph, user_graph, graph_filename="outputs/gateway.png")
                return

            if rest:
                # Run csp blocking on thread, run app in foreground
                block = False

            if block:
                # If blocking, run csp in foreground
                self.output = csp.run(
                    self.graph,
                    user_graph,
                    realtime=realtime,
                    starttime=starttime,
                    endtime=endtime or MAX_END_TIME,
                )
            else:
                # If not blocking, run csp on thread
                self.output = csp.run_on_thread(
                    self.graph,
                    user_graph,
                    realtime=realtime,
                    starttime=starttime,
                    endtime=endtime or MAX_END_TIME,
                )

            if rest:
                # these are temporary until a pydantic model is in place
                if isinstance(build_timeout, timedelta):
                    build_timeout = build_timeout.total_seconds()
                if isinstance(build_timeout, str):
                    build_timeout = int(build_timeout)

                # if graph shuts down or breaks, we need to shut down the web app
                # as well otherwise it "looks" like its running but its now
                self.web_app = self._build_web(ui=ui, timeout=build_timeout, _in_test=_in_test)
                log.info("Launching web server on:")
                url = f"http://{gethostname()}:{self.settings.PORT}"

                if ui:
                    if self.settings.AUTHENTICATE:
                        log.info(f"\tUI: {url}?token={self.settings.API_KEY}")
                    else:
                        log.info(f"\tUI: {url}")

                log.info(f"\tDocs: {url}/docs")
                log.info(f"\tDocs: {url}/redoc")

                # Run the web app
                # NOTE: this will block, except in test mode
                self.web_app.run(**uvicorn_kwargs)

            if _in_test:
                # return here, dont bother waiting
                return

            if rest:
                # send shutdown to csp if alive
                return self._shutdown()
            elif block:
                # wait for csp thread to be done
                return self._shutdown(wait=True)
            return self.output

        except KeyboardInterrupt:
            log.critical("Shutting down...(Keyboard Interrupt)")
            self._shutdown(user_initiated=True)
        except Exception:
            log.exception("Shutting down...")
            self._shutdown(user_initiated=False)
            # If we're here, we hit some form of error,
            # so re-raise it back to the caller
            raise

    def _get_web_app_class(self) -> Any:
        # FIXME ugly
        from csp_gateway.server import GatewayWebApp

        return GatewayWebApp

    def _build_web(self, ui: bool, timeout: int, _in_test: bool = False) -> Any:
        log.info(f"Building server with: {self.settings}")

        web_app_class = self._get_web_app_class()
        web_app = web_app_class(
            self,
            csp_thread=self.output,
            ui=ui,
            settings=self.settings,
            logger=log,
            _in_test=_in_test,
        )

        # Wait until graph is built by csp thread
        elapsed = 0

        while not self.running and not self.graph_build_failed:
            # FIXME ugly
            sleep(0.5)
            elapsed += 0.5

            if elapsed >= timeout:
                log.critical("Timeout during startup of graph, shutting down")
                raise RuntimeError("Graph start timeout")

            if not self.output.is_alive():
                log.critical("Graph start failure")
                raise RuntimeError("Graph start failure")

        if self.graph_build_failed:
            log.critical("Startup of graph failed, shutting down")
            raise RuntimeError("Graph build failure")

        # Revisit each module and connect to rest, if necessary
        for module in self.modules:
            if not module.disable:
                module.rest(web_app)

        web_app._finalize()

        return web_app

    def stop(self, **kwargs) -> Any:
        log.warning("Shutting down gateway...")

        if isinstance(self.output, csp.impl.wiring.threaded_runtime.ThreadRunner):
            # stop csp
            kwargs["user_initiated"] = True
            return self._shutdown(**kwargs)
        else:
            raise Exception("Gateway can only be stopped if started with `block=False`")

    def _shutdown(self, user_initiated: bool = True, wait: bool = False):
        # Now run through shutdown routine
        log.warning("Initiating shutdown...")

        if not isinstance(self.output, csp.impl.wiring.threaded_runtime.ThreadRunner):
            # nothing more to do
            return self.output

        if not wait:
            if self.output.is_alive():
                # First, try to stop the CSP thread if its alive
                try:
                    self.output.stop_engine()
                except Exception:
                    # If there was an error stopping it, we have an unclean shutdown
                    log.exception("CSP `stop_engine` exception")
                    raise

        # Now join the csp thread, allowing it to raise
        try:
            ret = self.output.join(suppress=False)
        except Exception:
            log.exception("CSP exception detected:")
            raise

        # Invoke shutdown for each of the modules in parallel
        # NOTE: Modules should shutdown independently of each other
        log.warning("Shutting down modules...")
        pool = multiprocessing.pool.ThreadPool()
        result = pool.map_async(lambda mod: mod.shutdown(), self.modules)
        # Wait for modules to shutdown safely, all modules should shutdown within this time
        try:
            result.get(timeout=self._module_shutdown_timeout)
        except multiprocessing.TimeoutError:
            log.warning(f"Shutting down modules took more than {self._module_shutdown_timeout}, forcefully shutting down...")

        # NOTE: Avoid having long running jobs in shutdown since python does not allow
        # forcefull termination of running threads. Therefore the threads will keep runnning
        # until they complete or the parent process dies
        pool.terminate()

        # Now run through shutdown routine
        log.warning("Shutting down webserver...")

        if ret is not None:
            # return value to caller
            log.warning("Returning value to user")
            return ret

        # No output from graph
        log.warning("No graph outputs detected, initiating hard shutdown")

        # else exit
        if not self._in_test:
            if user_initiated:
                log.warning("Shutting down webserver - CLEAN")

                # try clean
                os._exit(0)

                # should not get here
                sleep(5)

            else:
                log.warning("Shutting down webserver - UNCLEAN")

                # try exit
                os._exit(1)

                sleep(5)

            # NOTE: should be unreachable
            log.warning("Shutting down webserver - SIGTERM")

            # force sigkill to self
            os.kill(os.getpid(), signal.SIGKILL)

    def __hash__(self):
        # So that csp doesn't complain about inability to memoize
        return id(self)



from starlette.responses import Response
from starlette.staticfiles import StaticFiles
from starlette.types import Scope

__all__ = ("CacheControlledStaticFiles",)


class CacheControlledStaticFiles(StaticFiles):
    async def get_response(self, path: str, scope: Scope) -> Response:
        response = await super().get_response(path, scope)
        response.headers["Cache-Control"] = "public, max-age=604800"
        return response
